#ifndef _STRING_H_
#define _STRING_H_

#include <iostream.h>
#include <strstream.h>


class String
{
  public:

    String();

    String(const String& other);

    String(const char* newStr); // copy newStr into a new String

    String(char** strpp);	// *strpp is a malloc'd buffer now under
				// String's control.  *strpp will be NULL
				// on return.

    String(ostrstream& s);	// Set string to the contents of the 
				// string stream.  Freeze's stream by
				// calling s.str().

    String(int convert);	// converts an integer to a string

    String(long convert);	// converts an integer to a string

    String(double convert);	// contverts a floating point to a string

    ~String();

    int Length() const;		// characters in string

    String& Concat(const String& other); // put other at the end of this

    void LTrim();		// remove leading whitespace

    void RTrim();		// remove trailing whitespace

    void Trim();		// remove leading & trailing whitespace

    // Assignment operators:

    String& operator =(const String& other); // Set string to other

    String& operator =(const char* other); // Set string to other

    String& operator =(char** other);
				// Set string to other, which is already
				// malloc'd; now under control of String

    String& operator =(ostrstream& s);
				// Set string to the contents of the 
				// string stream.  Freeze's stream by
				// calling s.str().

    // Concatenation operators:

    String& operator +=( const String& other ); // add to this string

    // add this to other, but put it in a new string
    String operator +( const String& other ) const;

    // Comparison operators:
    int operator ==( const String& other ) const;
    int operator !=( const String& other ) const;
    int operator >=( const String& other ) const;
    int operator <=( const String& other ) const;
    int operator  >( const String& other ) const;
    int operator  <( const String& other ) const;

    // type conversions:
    operator const char*() const;
    operator long() const;
    operator double() const;

    friend ostream& operator <<(ostream& outs, const String& item);

  private:
    
    //-------------------------------------------------------------------------

    // This is the shared storage manager for strings.
    // It shares copies of it self with many strings via reference counting.
    // The String class is responsible for using this class properly.

    class Rep
      {
	public:

	  Rep();		// create an empty string

	  Rep(const Rep& other); // deep copy

	  Rep(const char* newStr); // copies newStr

	  Rep(char** strpp);	// *strpp is malloc'd buffer now under
					// Rep's control.
					//  *strpp will be NULL on return.

	  ~Rep();		// cleans up the representation.
				// this should only be called when 
				// ref_count == 0.

	  void Resize(int newLen);	// make the string large enough to hold
					// at least newLen characters.
					// Truncate string if newLen < length

	  Rep* AddReference();	// add a reference to this rep
					// return this

	  int DeleteReference();	// remove a reference to this rep
					// return count

	  void Init();		// set initial variable values.
				// called by all constructors.

	  Rep& operator =(const Rep& other); // n/a
	  
	  int ref_count;		// number of Strings sharing this rep

	  int size;			// amount of allocated space

	  char* str;			// pointer to string

      };

    //------------------------------------------------------------------------


    Rep* rep;

    void UniqueRep();		// get our own copy of rep

    void DeleteRep();		// decrement ref count & delete rep if needed

};


#include "String.i"

#endif // _STRING_H_

