// *************************************************************************
//    IncrLoader.C
// *************************************************************************

#ifdef INCRLOADER

#include <sys/types.h>
#include <unistd.h>

#include <fcntl.h>
#include <sys/exec.h>

    // Where is this supposed to be defined?
/*  #define    PAGSIZ 1024  */

#include "IncrLoader.h"
#include "Globals.h"
#include "TupleIterator.h"
#include "Identify.h"

// *************************************************************************

#define DUMMY_FN_NAME "_CORAL_dummy"
#define CORAL_FILENAME_TEMPFILE  "_fname_coral.temp"

extern char *coral_name;    // declared in main.C

// *************************************************************************

#ifndef __GNUC__
#ifndef HPUX
extern "C"  char *sbrk(int);
extern "C"  read(int, char *, int);
#endif
extern "C"  close(int);
#endif

// *************************************************************************
// move this to Config.h

#ifdef __GNUC__
#    define _CORAL_COMPILER   "gcc"
#    define _INCR_LOAD_OPTS   "-c -w"
#    define _CORAL_LOADER     "ld"
#else
#    define _CORAL_COMPILER   "CC"
#    define _CORAL_LOADER     "ld"
#    ifdef MIPS
#        define _INCR_LOAD_OPTS   "-c -w -G 0"
#    else
#        define _INCR_LOAD_OPTS   "-c"
#    endif
#endif

// *************************************************************************

class exec_function {
public:
    caddr_t   func;
    exec_function *next;
    exec_function() { next = NULL; }
};

// *************************************************************************
static exec_function * incr_load(const char *fname)
{
    char filename[200];
    strcpy(filename, fname);

    char buf[1000] ;
    int fd ,len ;
    len = strlen(filename);

    // first determine the full pathname of this program !
    sprintf(buf, "which %s 2>/dev/null > %s", coral_name,
        CORAL_FILENAME_TEMPFILE);

    if (system(buf)) {
        *(Cor_exEnv->err_strm)
            << "unable to determine pathname of executing program :";
        *(Cor_exEnv->err_strm) << " assuming " <<  coral_name << "\n";
    } else {
        ostream *fd;
        if (!(fd = Cor_open_infile(CORAL_FILENAME_TEMPFILE))) {
            *(Cor_exEnv->err_strm)
                << "unable to determine name of executing program :";
            *(Cor_exEnv->err_strm) << " assuming " <<  coral_name << "\n";
            unlink(CORAL_FILENAME_TEMPFILE);
        }
        else {
            fd->gets(coral_name, 200);
            // the last character is a \n, and needs to be overwritten
            coral_name[strlen(coral_name)-1] = '\0';
            delete fd;
            unlink(CORAL_FILENAME_TEMPFILE);
        }
    }

    // Check that the file is a .o file
    if (filename[len-2] != '.') {
        *(Cor_exEnv->err_strm)
            << "Input file should be a .C file, or a .S file\n";
#        ifdef MIPS
            *(Cor_exEnv->err_strm)
                << "or a .o file compiled with the -G 0 option\n";
#        else
            *(Cor_exEnv->err_strm) << "or a .o file\n";
#        endif
        return NULL;
    }

    if ((filename[len-1] != 'o') || (filename[len-2] != '.')) {
        if ((filename[len-1] != 'C') && (filename[len-1] != 'S')) {
            *(Cor_exEnv->err_strm)
                << "Input file should be a .C file, or a .S file\n";
#            ifdef MIPS
                *(Cor_exEnv->err_strm)
                    << "or a .o file compiled with the -G 0 option\n";
#            else
                *(Cor_exEnv->err_strm) << "or a .o file\n";
#            endif
            return NULL;
        }
        if (filename[len-1] == 'S') {
            sprintf(buf, "translator -i %s -o %s", filename, filename);
            buf[strlen(buf)-1] = 'C';
            if (system(buf)) {
                *(Cor_exEnv->err_strm)
                    << "Could not run the translator on " << filename << endl;
                return NULL;
            }
            filename[len-1] = 'C' ;
        }    

        sprintf(buf, "%s %s -I%s/includes %s", _CORAL_COMPILER,
            _INCR_LOAD_OPTS, Cor_exEnv->coral_path, filename);
        if (system(buf)) {
            *(Cor_exEnv->err_strm) << "Could not compile " <<  filename << "\n";
            return NULL;
        }

#        ifdef DEBUG
            *(Cor_exEnv->err_strm) << "Compiled file " <<  filename << "\n";
#        endif
        filename[len-1] = 'o';
    }

    ostream *fdes ;

    //    Create the loadable code with all references patched
    caddr_t oadx = (caddr_t)sbrk(0);
#    ifdef MIPS
        caddr_t adx = ((char*)oadx) + NBPG - (((int)oadx) % NBPG);
#    else
        /** SUN **/
        caddr_t adx = ((char*)oadx) + PAGSIZ - (((int)oadx) % PAGSIZ);
#    endif

    sprintf(buf, "%s -N -T %X -A %s %s -o %s.new", _CORAL_LOADER, 
        adx, coral_name, filename, filename);
    if (system(buf)) {
        *(Cor_exEnv->err_strm) << "Failed to load\n";
        sprintf(buf, "%s.new", filename);
        unlink(buf);
        return NULL ;
    }

#    ifdef DEBUG
        *(Cor_exEnv->err_strm)
            << "created file " << filename
            << ".new with references resolved\n";
#    endif

    sprintf(buf, "%s.new", filename);
    strcpy(filename, buf);

    //    Perform the incremental load
    caddr_t ldadx;

#    ifdef MIPS
        fd = open(buf, O_RDONLY, 0);

        struct filehdr Filehdr;
        struct aouthdr Aouthdr;
        read(fd, (char *)&Filehdr, sizeof(struct filehdr));
        read(fd, (char *)&Aouthdr, sizeof(struct aouthdr));
        
        sbrk(NBPG - (((int)oadx) % NBPG));

        ldadx = (caddr_t)sbrk((int)(Aouthdr.tsize + 
                        Aouthdr.dsize + Aouthdr.bsize));


        lseek(fd, (int)(N_TXTOFF(Filehdr, Aouthdr)), SEEK_SET);
        read(fd, ldadx, ((int)(Aouthdr.tsize + Aouthdr.dsize + Aouthdr.bsize)));
#    endif // MIPS

#    ifdef SUN
        fd = open(buf, O_RDONLY);

        struct exec Exec;
        read(fd, (char *)&Exec, sizeof(struct exec));
        sbrk(PAGSIZ - (((int)oadx) % PAGSIZ));

        ldadx = (caddr_t)sbrk((int)(Exec.a_text + Exec.a_data + Exec.a_bss));
        read(fd, ldadx, ((int)(Exec.a_text + Exec.a_data)));
#    endif

    close(fd);

#    ifdef DEBUG
        *(Cor_exEnv->err_strm) << "load addr is " << ((void*) ldadx) << endl;
#    endif

    // retrieve the relocated address of the extern function

    sprintf(buf, "nm %s | grep %s > _CORAL.load_tmp2", filename, DUMMY_FN_NAME);
    if (system(buf)) {
        *(Cor_exEnv->err_strm)
            << "Unable to locate installing function "
            << DUMMY_FN_NAME << " in relocated code\n";
        unlink(filename);
        unlink("_CORAL.load_tmp2");
        delete ldadx;
        return NULL ;
    }

    if (!(fdes = Cor_open_infile("_CORAL.load_tmp2"))) {
        *(Cor_exEnv->err_strm)
            << "Unable to open temporary file _CORAL.load_tmp2\n";
        unlink(filename);
        unlink("_CORAL.load_tmp2");
        delete ldadx;
        return NULL;
    }

    // retrieve the address of the extern function
    int return_address;
    char temp_char;
    exec_function *func_list = new exec_function;

    if (fscanf(fdes, "%X %c %s", &return_address, &temp_char, buf) != 3) {
        *(Cor_exEnv->err_strm)
            << "Unable to recover relocated installing function address\n";
        delete fdes;
        unlink(filename);
        unlink("_CORAL.load_tmp2");
        delete ldadx;
        return func_list;
    }

    exec_function *temp; int count    = 0;
    temp = func_list->next;
    func_list->next = new exec_function;
    func_list->next->func = (caddr_t)return_address;
    func_list->next->next = temp;
    count++;
    
    while (fscanf(fdes, "%X T %s", &return_address, buf) == 2) {
        temp = func_list->next;
        func_list->next = new exec_function;
        func_list->next->func = (caddr_t)return_address;
        func_list->next->next = temp;
        count++;
    }

    delete fdes;
    unlink(filename);
    unlink("_CORAL.load_tmp2");

    return func_list;
}

// *************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return;}

void loadIncrFile(char *filename, Cor_TupleIterator& iterator)
{
    typedef void (*PF)(...);
    
    exec_function *func_list = incr_load(filename);
    if (!func_list || !(func_list->next)) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- incremental load failure\n";
        delete [] filename;
        FAIL_RETURN;
    }
    while (func_list->next) {
        PF anewfunc = (PF) func_list->next->func;
        (*anewfunc)();
        exec_function *temp = func_list->next;
        delete func_list;
        func_list = temp;
    }
}

#undef FAIL_RETURN

// *************************************************************************

#endif INCRLOADER

