// *************************************************************************
//    Trail.h
// *************************************************************************

#ifndef _Trail_H_
#define _Trail_H_

#include "BindEnv.h"
#include "Term.h"
#include "Debug.h"

// *************************************************************************

struct TrailElement {
    Cor_BindEnv *bindenv;
    Variable     var;
};

// *************************************************************************
// The Trail is a means of allowing environments to be created (during
// rule evaluation) and then discarded when they are not needed any more.
// It is a stack of environments (BindEnv/ Variable) pairs which can
// be pushed to grow it, and popped back to a fixed StackMark, to bring
// the environment back to its old state.

class Cor_Trail {
    TrailElement *Stack;
    int Limit;

  public:
    // Size of current Trail stack.
    int TrailSize;

    // Creates a trail with the specified initial sizes.
    Cor_Trail(int trailsize);
   ~Cor_Trail();

    // Return specified element of trail
    inline TrailElement& operator [](int i)
    {
        return (Stack[i]);
    }

    // Pop trail to specified size. Clear (in BindArray) variables that are 
    // found on the popped portion of the trail.
    inline void PopTrail(int new_size)
    {
#ifdef    DO_CALL_STACK
        CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
        while (TrailSize > new_size) {
            TrailSize--;
            if (Stack[TrailSize].var >= 0) {
                (Stack[TrailSize]).bindenv->clear(Stack[TrailSize].var);
            } else {
                delete Stack[TrailSize].bindenv;
            }
        }
    }
                
    inline void PushTrail(Cor_BindEnv *env, Variable var )
    {
#ifdef    DO_CALL_STACK
        CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
        if (TrailSize >= Limit) {
            if (Limit == 0) {
                  Limit = 255;
                  Stack = new TrailElement[Limit];
            } else {
                  TrailElement *tmpStack = Stack;
                  int len = Limit;
                  Limit += Limit;
                  Stack = new TrailElement[Limit];
                  for (int i=0; i<len; i++)
                      Stack[i] = tmpStack[i];
                  delete [] tmpStack;
            }
        }
        Stack[TrailSize].bindenv = env;
        Stack[TrailSize].var = var;
        TrailSize++;
    }

    // Push a bindenv on trail; 
    // The argument -1 indicates the whole bindenv is to be 
    // deleted on backtracking.
    inline void PushTrail(Cor_BindEnv *env)
    {
        PushTrail(env, -1);
    }

    // For debugging, check that this BindEnv is not present
    // in the trail.
    int is_InTrail(Cor_BindEnv *env) {
        int j;

        if (this) 
            for (j=0; j < TrailSize; j++) {
                if (env == Stack[j].bindenv)
                    return(1);
            }

        return(0);
    }
};

// *************************************************************************

extern Cor_Trail *Cor_CurTrail;

// *************************************************************************

#endif _Trail_H_

