
//**************************************************************************
//    ConnectArray.h
//
//    The ConnectArray Class - An abstract class for a general array of
//    connections.
//
//**************************************************************************

#ifndef    _CONNECTARRAY_H
#define    _CONNECTARRAY_H

#include "Connection.h"

//**************************************************************************

                                // The maximum number of physical 
                                // connections allowed.
#define    MAX_CONNECTIONS        64

//**************************************************************************

class Cor_ConnectArray {

  public:  
    
    int            cur_conn;    // The current connection 

    int            num_conn;    // The current number of connections

                                // The maximum number of physical connections
                                // allowed (at once) by this ConnectArray
                                // This can be changed dynamically
                                // (maybe to control load)
    int            max_conn;      


    CoralStream      *expl_serv_strm;   // The server connection (for explain)

    Cor_Connection   *connArray[MAX_CONNECTIONS];


    virtual ~Cor_ConnectArray();

      // Is this ConnectArray local (a common question)
    virtual int is_local() { return(0); }

      // Return the current number of connections
    inline int NumConnections() { return(num_conn); }

      // Return the maximum number of connections
    inline int MaxConnections() { return(max_conn); }
    
      // Set the maximum number of connections (must be <= MAX_CONNECTIONS)
    inline void SetMaxConnections(int i) {
        max_conn = (i < MAX_CONNECTIONS) ? i : MAX_CONNECTIONS;
    }

      // Delete a connection
    virtual void DeleteConnection(int conn_indx) = 0;

      // Establish a connection for Explain
    virtual int ExplainConnection() = 0;

      // Get the next connection with data and set it up to run
    virtual int NextConn() = 0;

      // Switch to this connection (switches to current user there)
    inline int  useConnection(int conn_id) {
        return( (connArray[conn_id])->useConnection() );
    }

      // Return a pointer to the current connection
    inline Cor_Connection *currentConn() {
        return(connArray[cur_conn]);
    }

      // List users for the connection specified, if conn is Null, list all
      // users on all connections
    virtual int ListWho(CoralStream *strm, Cor_Connection *conn = NULL);

      // Set up a temporary Local connection. 
    int LocalConnection();

      // and delete it.
    void DeleteLocalConnection();
};

// *************************************************************************

#endif    _CONNECTARRAY_H
