// **************************************************************************
//    Interface.C
// **************************************************************************

#include <stdio.h>

#include "Globals.h"
#include "ParseUtils.h"

#include "Interface.h"
#include "HashSimpleRelation.h"
#include "BuiltinUserRelRelation.h"
#include "BuiltinUserTupleRelation.h"
#include "Util.h"
#include "Debug.h"

// **************************************************************************
Cor_CArg *make_arg (char *name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *ret = new Cor_CArg(name);

    return(ret);
}

// **************************************************************************
Cor_CArg *make_arg (Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *ret = new Cor_CArg(name);

    return(ret);
}

// **************************************************************************
Cor_CArg *make_var (int i)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *carg = new Cor_CArg();
    carg->set_var(i);

    return(carg);
}

// **************************************************************************
Cor_CArg *make_var (char *, int i)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *carg = new Cor_CArg();
    carg->set_var(i);

    return(carg);
}

// **************************************************************************
Cor_CArg *make_var (Cor_CArg&, int i)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *carg = new Cor_CArg();
    carg->set_var(i);

    return(carg);
}

// **************************************************************************
Cor_CArg *make_arg (char *func_name, Cor_ArgList *args)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg name_arg = Cor_CArg(func_name);
    Cor_CArg *fa = new Cor_CArg(name_arg, args->count());

    for (int i=0; i < args->count(); i++) {
        (fa->args())[i] = (*args)[i];
    }

    return(fa);
}

// **************************************************************************
Cor_CArg *make_cons(Cor_CArg *car, Cor_CArg *cdr)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *carg = new Cor_CArg(car,cdr);
    return(carg);
}

// **************************************************************************
Cor_CArg *make_nil()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *carg = new Cor_CArg(); 
    carg->set_kind(NIL_ARG);

    return(carg);
}

// **************************************************************************
// Return the length of a [...] type list, or < 0 if it is not a list.
//
int list_length( Cor_CArg *list )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (! list->is_list())
        return(-1);

    if (list->is_nil())
        return(0);

    return(list->num_items() - 1);
}

// **************************************************************************
// Make a list of the arguments given i.e (X,1,[A,B],M) -> [X,1,[A,B],M]
//
Cor_CArg *make_list(int len, ...)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    va_list ap;
    va_start(ap,len);
   
    Cor_CArgStack carg_array;

    Cor_CArg *list = NULL;
    Cor_CArg *carg;

    for (i=0; i < len; i++) {
        carg = va_arg(ap, Cor_CArg*);
        carg_array.push(carg);
    }

    while ((carg = carg_array.pop())) {
        list = make_cons(carg,list); 
    }

     
    va_end(ap);

    if (!list)
        list = make_nil();

    return(list);
}

// **************************************************************************
Cor_ArgList *make_arglist (int len, ...)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i=0;
    va_list ap;
    va_start(ap,len);

    Cor_ArgList *al = new Cor_ArgList(len);

    for(i =0; i < len; i++) {
        (*al)[i] = *(va_arg(ap, Cor_CArg*));
    }

    va_end(ap);
    return al;
}

// **************************************************************************
//  Version that destroys the passed in CArgs
Cor_ArgList *make_arglist_d (int len, ...)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i=0;
    va_list ap;
    va_start(ap,len);

    Cor_ArgList *al = new Cor_ArgList(len);

    for(i =0; i < len; i++) {
        Cor_CArg *carg = va_arg(ap, Cor_CArg*);
        (*al)[i] = *carg;
        ASSERT((carg != NULL));
        carg->destroy();
    }

    va_end(ap);
    return al;
}

// **************************************************************************
Cor_ArgList *make_arglist (const Cor_CArg& a1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList *alist = new Cor_ArgList(1);
    (*alist)[0] = a1;
    return alist;
}

// **************************************************************************
Cor_ArgList *make_arglist (const Cor_CArg& a1, const Cor_CArg& a2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList *alist = new Cor_ArgList(2);
    (*alist)[0] = a1;
    (*alist)[1] = a2;
    return alist;
}

// **************************************************************************
Cor_ArgList *make_arglist (const Cor_CArg& a1,
            const Cor_CArg& a2, const Cor_CArg& a3)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList *alist = new Cor_ArgList(3);
    (*alist)[0] = a1;
    (*alist)[1] = a2;
    (*alist)[2] = a3;
    return alist;
}

// **************************************************************************
Cor_ArgList *concat_arglists (Cor_ArgList *al1, Cor_ArgList *al2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int c1 = al1->count();
    int c2 = al2->count();
    Cor_ArgList *alist = new Cor_ArgList(c1+c2);

    for (int i=0; i < c1; i++){
        (*alist)[i] = (*al1)[i];
    }

    for (int j=0; j < c2; j++){
        (*alist)[c1+j] = (*al2)[j];
    }
    return alist;
}

// **************************************************************************
Cor_ArgList *make_vararglist (int n)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList *alist = new Cor_ArgList(n);
    for (int i=0; i<n; i++) {
        Cor_CArg *carg = make_var(i);
        (*alist)[i] = *carg;
        carg->destroy();
    }
    return(alist);
}

// **************************************************************************
int is_num (const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(a.is_num());
}

// **************************************************************************
int is_constant (const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // assuming only two types of constants
    return (a.is_symbol() || a.is_num());
}

// **************************************************************************
int is_var (const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(a.is_var());
}

// **************************************************************************
int is_functor (const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(a.is_functor());
}

// **************************************************************************
int is_list (const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(a.is_list());
}

// **************************************************************************
int is_nil (const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(a.is_nil());
}

// **************************************************************************
int make_int(const Cor_CArg& a)
{ 
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return a.NumIval(); 
}

// **************************************************************************
short make_short(const Cor_CArg& a)
{ 
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (short)a.NumIval();
}

// **************************************************************************
long make_long(const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (long)a.NumIval();
}

// **************************************************************************
float make_float(const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return a.NumDval();
}

// **************************************************************************
double make_double(const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return a.NumDval();
}

// **************************************************************************
char *make_string(const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *result = new char[a.symlen() + 1];
    memcpy(result, a.symbol(), (size_t) a.symlen()+1);
    return(result);
}

// **************************************************************************
void *make_chunk(Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(a.chunk());
}

// **************************************************************************
char *functor_name(const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *result = new char[a.symlen() + 1];
    memcpy(result, a.symbol(), (size_t) a.symlen() +1);
    return(result);
}

// **************************************************************************
Cor_ArgList *functor_args (Cor_CArg *a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return ( &(a->args()) );
}

// **************************************************************************
Cor_CArg *make_car (const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *carg = a.car();

    if (!carg)
        carg = make_nil();
    else
        carg = carg->dup();

    return(carg);
}

// **************************************************************************
Cor_CArg *make_cdr (const Cor_CArg& a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(a.cdr());
}

// **************************************************************************
Cor_Tuple *make_tuple(int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(new Cor_Tuple(arity));
}

// **************************************************************************
Cor_Tuple *make_tuple(Cor_ArgList *arglist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(new Cor_Tuple(arglist));
}

// **************************************************************************
Cor_Relation *make_rel(char *rel_name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg rel_arg = Cor_CArg(rel_name);
    return Cor_make_relation(rel_arg, arity);
}

// **************************************************************************
int call_coral(char *exp_pred_name, Cor_Tuple *query, Cor_Relation *&result)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg pred = Cor_CArg(exp_pred_name);
    Cor_Relation *rel = Cor_find_external_relation(pred, query->args().count());
    if (!rel) {
        *(Cor_exEnv->err_strm)
            << "Error: call_coral: Relation " << exp_pred_name
            << " not found\n";
        return -1;
    }

    if (!result)
        result = new Cor_HashSimpleRelation(rel->arity());
    else {
        if (rel->arity() != result->arity()) {
            *(Cor_exEnv->err_strm)
                 << "Error: call_coral: arity mismatch "
                 << "between query and result relation\n";
            return -1;
        }
    }
        

    Cor_BindEnv env(query->args().max_vars());

    Cor_TupleIterator it(rel, query->args(), &env, NULL, NULL);
    int answer_count = 0;
#    ifdef DO_TRAIL
        Cor_StackMark stack_mark;
#    endif
    for (;;) {
        it.get_next_tuple();
        if (it.no_match())
            break;
        else {
            result->insert_new(query->args(), &env);
            answer_count ++;
        }
#        ifdef DO_TRAIL
            stack_mark.pop_to();
#        endif
    }
    return answer_count;
}

// **************************************************************************
/* this function is probably outdated --- PRAVEEN */

int call_coral(char *exp_pred_name, Cor_Relation *queries,
            Cor_Relation *&result)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg pred = Cor_CArg(exp_pred_name);
    Cor_Relation *rel = Cor_find_external_relation(pred, queries->arity());
    if (!rel) {
        *(Cor_exEnv->err_strm) << "Error: call_coral: Relation "
            << exp_pred_name << " not found\n";
        return -1;
    }

    if (!result)
        result = new Cor_HashSimpleRelation(rel->arity());
    else {
        if (rel->arity() != result->arity()) {
            *(Cor_exEnv->err_strm)
                 << "Error: call_coral: arity mismatch "
                 << "between query and result relation\n";
            return -1;
        }
    }

    int answer_count = 0;

    Cor_BindEnv env0(queries->arity());
    Cor_ArgList *arglist0 = make_vararglist(queries->arity());
    // create an arglist with vars 0 through (arity-1) for it.

    Cor_TupleIterator it0(queries,*arglist0, &env0, NULL, NULL);
            /** to iterate over tuples in queries **/

#    ifdef DO_TRAIL
        Cor_StackMark stack_mark0;
#    endif
    for (;;) {
        it0.get_next();
        if (it0.no_match())
            break;
#        ifdef DO_TRAIL
            Cor_StackMark stack_mark1;
#        endif
        Cor_TupleIterator it1(rel, *arglist0, &env0, NULL, NULL);

        for(;;) {
            it1.get_next();
            if (it1.no_match())
                break;
            else {
                result->insert_new(*arglist0, &env0);
                answer_count ++;
            }
#           ifdef DO_TRAIL
            stack_mark1.pop_to();
#           endif
        }
#       ifdef DO_TRAIL
        stack_mark0.pop_to();
#       endif
    }

#   ifdef DO_TRAIL
    stack_mark0.pop_to();
#   endif

    return answer_count;
}

// **************************************************************************
//    add_builtin is only used from BuiltinRelation.C. It is NOT part of the
//    C++ interface

int add_builtin(char *pred, int arity, UserRelProc pred_solver)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *tmp = new Cor_CArg(pred);

    Cor_BuiltinUserRelRelation *dummy = new Cor_BuiltinUserRelRelation (arity, 
        tmp, pred_solver);

    if (dummy)    // This is here to get rid of the "unused" warning
        ;    // Nothing

    delete tmp;
    return(1); // Success
}

// **************************************************************************
int add_builtin(char *pred, int arity, UserTupleProc pred_solver)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *tmp = new Cor_CArg(pred);

    Cor_BuiltinUserTupleRelation *dummy =
      new Cor_BuiltinUserTupleRelation(arity,tmp,pred_solver);

    if (dummy)    // This is here to get rid of the "unused" warning
        ;    // Nothing

    delete tmp;

    return(1); // Success
}

// **************************************************************************
// execute the i'th ParserStruct in the ExecutionList
int Cor_ExecutionList::execute(int i, int& param_index)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList *al;

    if (i >= actual_no_of_terms)
        return(1);

      // Need to fill in the parameters here
    if ((param_index < no_of_params) && (! (exec_terms[i])->no_param_args)) {

        if ( (exec_terms[i])->cur_rule->head )
            al = &((exec_terms[i])->cur_rule->head->args);
        else
            al = NULL;

        if ( ((exec_terms[i])->orig_args == NULL)
                  && al && (al->arity() != 0) ) {
            (exec_terms[i])->orig_args = al->dup();
        }

        Cor_ArgList *orig_al = (exec_terms[i])->orig_args;

        if (orig_al) {

            int cnt = 0;
            int len = al->arity();

            for (int j = 0; j < len; j++ ) {

                (*al)[j] = (*orig_al)[j];
                int tmp = (*al)[j].substitute_params(&params);
                cnt += tmp;
                param_index += tmp;
            }

            if (cnt == 0)
                (exec_terms[i])->no_param_args = 1;
        }
    }

    ASSERT((param_index <= no_of_params));

    (exec_terms[i])->execute();

    return 0;    
}

// **************************************************************************
int Cor_ExecutionList::execute()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int param_index = 0;

    for (int i = 0; (i < actual_no_of_terms); i++) {
        execute(i, param_index);
    }    

    return 0;
}

// **************************************************************************
void Cor_ExecutionList::add(Cor_ParserStruct *pstruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (actual_no_of_terms >= max_no_of_terms) {
        Cor_ParserStruct **temp = exec_terms;
        max_no_of_terms = max_no_of_terms + step_size;
        exec_terms = new Cor_ParserStruct*[max_no_of_terms];
        for (int i = 0; i < actual_no_of_terms; i++)
            exec_terms[i] = temp[i];        
    }
    exec_terms[actual_no_of_terms++] = pstruct;
}

// **************************************************************************
// process the top item on the input stack

void Cor_processInputStack()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_exEnv->C_exec_mode = COR_DOING_INTERPRET;

    if (! Cor_parseStack->is_empty())
        Cor_parseStack->parse(Cor_exEnv->err_strm,1);

    Cor_exEnv->C_exec_mode = COR_DOING_IMPERATIVE;
}

// **************************************************************************
Cor_ScanDesc::Cor_ScanDesc(Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!rel)
        return;

    stackmark = new Cor_StackMark();

    env = new Cor_BindEnv(rel->arity());
    arg_list = make_vararglist(rel->arity());
    it = new Cor_TupleIterator(rel, *arg_list, env);
}

// **************************************************************************
Cor_ScanDesc::Cor_ScanDesc(Cor_Relation *rel, Cor_Tuple *match_tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    extern int Cor_num_vnames;

    if (!rel || !match_tuple)
        return;

    stackmark = new Cor_StackMark();
    arg_list = NULL;

    if (rel->arity() > Cor_num_vnames)
        env = new Cor_BindEnv(rel->arity());
    else
        env = new Cor_BindEnv(Cor_num_vnames);

    it = new Cor_TupleIterator(rel, match_tuple->args(), env);
}

// **************************************************************************
Cor_Tuple *Cor_ScanDesc::next_tuple()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!this)
        return(NULL);

    if (stackmark)
        stackmark->pop_to();

    env->clear_all();

    Cor_Tuple *ret_tuple = it->get_next_tuple();

    // Note :: pipelining does not return an actual tuple. instead it merely
    //    provides bindings for variables. Since the interface expects
    //    a handle on a real tuple, we have to materialize it
    if (!it->no_match() && !ret_tuple) {
        for (int i = 0; i < it->arg_list.arity(); i++) {
            Cor_TermLink *rv = NULL;
            Cor_CArg tmp;
            it->arg_list[i].simplify(tmp, it->bindenv, rv, it->bindenv, NULL);
        }
        ret_tuple = new Cor_Tuple(it->arg_list);
    }
    return ret_tuple;
}

// **************************************************************************
Cor_ScanDesc::~Cor_ScanDesc()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!this)
        return;

    if (stackmark) {
        stackmark->pop_to();
        delete stackmark;
    }

    if (arg_list)
        delete arg_list;

    delete it;
    delete env;
}

// **************************************************************************
Cor_ScanDesc *open_scan(char *example_fact)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ScanDesc *sd = NULL;
    Cor_Relation *rel;
    Cor_Tuple    *tuple;
    Cor_CArg      rel_nme;
    int           arity;

    extern Cor_CArg  *Cor_RelName;

      // In case we previously parsed a table file
    if (Cor_RelName) {
        Cor_RelName->destroy();
        Cor_RelName = NULL;
    }

    tuple = Cor_parseFact(example_fact, rel_nme, &arity); 

    if (tuple == NULL) {
#ifdef    DEBUG
        cout << "parseFact failed on " << example_fact << endl;
#endif
        return(NULL);
    }

    rel = find_relation(rel_nme, arity);

    if (rel)
        sd = new Cor_ScanDesc(rel, tuple);
#ifdef    DEBUG
    else
        cout << "find_relation failed on " << rel_nme << endl;
#endif

    return(sd);
}

// **************************************************************************
