// *************************************************************************
//    Stack.C
// *************************************************************************

#include <stdlib.h>

#include "Stack.h"
#include "Globals.h"
#include "Debug.h"

        // The stack can grow, but start at a reasonable size
#define    DEFAULT_SIZE    8

// *************************************************************************
int  Cor_IntStack::push(int x)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (length >= data_size) {

        if (data_size == 0) {
            data_size = DEFAULT_SIZE;
            data = new int[data_size];
        } else {

            int len = data_size;
            data_size += data_size;
            int *tmp_data = data;
            data = new int[data_size];
            memcpy(data, tmp_data, len * sizeof(int));
            delete [] tmp_data;
        }
    }

    data[length++] = x;

    return(length-1);
}

// *************************************************************************
void Cor_IntStack::insert_at(int pos, int item)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Defensive programming :-)
    if (pos > length)
        return;

      // This ensures there is room available
      // (As a side-effect, it increments the length)
    push(0);

    for (int i = length-1; i > pos; i--)
        data[i] = data[i-1];

    data[pos] = item;
}

// *************************************************************************
//  copies contents of stack stk to this one
void Cor_IntStack::copy(Cor_IntStack& stk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] data;
    length = stk.length;
    data_size = stk.data_size;
    data = new int[data_size];
    for (int i = 0; i < length; i++)
        data[i] = stk.data[i];
}

// *************************************************************************
// Stream printing
ostream& operator<< (ostream& out, const Cor_IntStack& ilist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (ilist.length)
        out << ilist.data[0];
 
    for (int j=1; j < ilist.length; j++)
        out << ", " << ilist.data[j];

    return(out);
}

// *************************************************************************
int  Cor_VoidStack::push(void *x)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (length >= data_size) {

        if (data_size == 0) {
            data_size = DEFAULT_SIZE;
            data = new (void *)[data_size];
        } else {
            int len = data_size;
            data_size += data_size;
            void **tmp_data = data;
            data = new (void*)[data_size];

            memcpy(data, tmp_data, len * sizeof(void*));
            delete [] tmp_data;
        }
    }

    data[length++] = x;

    return(length-1);
}

// *************************************************************************
void Cor_VoidStack::insert_at(int pos, void *item)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Defensive programming :-)
    if (pos > length)
        return;

      // This ensures there is room available
      // (As a side-effect, it increments the length)
    push(NULL);

    for (int i = length-1; i > pos; i--)
        data[i] = data[i-1];

    data[pos] = item;
}

// *************************************************************************
//  copies contents of stack stk to this one
void Cor_VoidStack::copy(Cor_VoidStack& stk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] data;
    length = stk.length;
    data_size = stk.data_size;
    data = new (void *)[data_size];
    for (int i = 0; i < length; i++)
        data[i] = stk.data[i];
}

// *************************************************************************
int  Cor_CArgStack::push(Cor_CArg *x)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (length >= data_size) {
        if (data_size == 0) {
            data_size = DEFAULT_SIZE;
            data = new Cor_CArg*[data_size];
        } else {
            int len = data_size;
            data_size += data_size;
            Cor_CArg **tmp_data = data;
            data = new Cor_CArg*[data_size];

            memcpy(data, tmp_data, len * sizeof(Cor_CArg*));
            delete [] tmp_data;
        }
    }

    data[length++] = x;

    return(length-1);
}

// *************************************************************************
void Cor_CArgStack::insert_at(int pos, Cor_CArg *item)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Defensive programming :-)
    if (pos > length)
        return;

      // This ensures there is room available
      // (As a side-effect, it increments the length)
    push(NULL);

    for (int i = length-1; i > pos; i--)
        data[i] = data[i-1];

    data[pos] = item;
}

// *************************************************************************
//  copies contents of stack stk to this one
void Cor_CArgStack::copy(Cor_CArgStack& stk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    empty();
    delete [] data;

    ref_cnt = 0;
    length = stk.length;
    data_size = stk.data_size;
    delete_cargs = 1;    // Since we're dup'ing them all

    data = new (Cor_CArg *)[data_size];

    for (int i = 0; i < length; i++)
        data[i] = stk.data[i]->dup();
}

// *************************************************************************
// Copy the Stack as far as specified (used when storing lists)
void Cor_CArgStack::copy_bottom(Cor_CArgStack& stk, int num_items)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] data;
    ref_cnt = 0;
    length = num_items;
    data_size = stk.data_size;

    data = new (Cor_CArg *)[data_size];

    for (int i = 0; i < length; i++)
        data[i] = stk.data[i]->dup();
}
// *************************************************************************
//  Build an arglist from the current contents of the stack
void Cor_CArgStack::build_arglist(Cor_ArgList& args)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this);

    int cnt = count();
    args.set_count(cnt);

    for (int j=0; j < cnt; j++) {

        Cor_CArg *carg = data[j];

        args[j] = *carg;

          // No one else should use this pointer
        data[j] = NULL;
        delete carg;
    }

    length = 0; 
}

// *************************************************************************
//  Empty out the stack - delete each element
void Cor_CArgStack::empty()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    for (int j=0; j < length; j++) {

        delete data[j];

          // No one else should use this pointer
        data[j] = NULL;
    }

    length = 0; 
}

// *************************************************************************
// Stream printing
ostream& operator<< (ostream& out, const Cor_CArgStack& ca_list)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (ca_list.length)
        out << *(ca_list.data[0]);
 
    for (int j=1; j < ca_list.length; j++)
        out << ", " << *(ca_list.data[j]);

    return(out);
}

#undef    DEFAULT_SIZE
// *************************************************************************
