
%{
/*****************************************************************************/
/*    Scanner.l    */
/*****************************************************************************/

#define    YY_NO_UNPUT
#define    YY_SKIP_YYWRAP

//extern "C" int yywrap();
int yywrap();

#include <iomanip.h>

#include "Globals.h"
#include "Scanner.h"
#include "Gram.tab.h"

#ifndef MAX_STR_LEN
#define MAX_STR_LEN    1024
#endif

static char  str_buff[MAX_STR_LEN];
static char *str_end = &str_buff[MAX_STR_LEN-1];
static char *str_ptr;

extern int   Cor_reading_table;

#undef YY_INPUT

#    ifdef STRING_PARSE

#define    min(X,Y)    ((X)<(Y) ? (X) : (Y))

    // Scan input from a string (or an array of strings) instead
#define YY_INPUT(buf,result,max_size) \
  { \
     int n = min(max_size, Cor_parseEnv->parselimit - Cor_parseEnv->parseptr);\
     if (n>0){ \
	bzero(buf, max_size); \
        memcpy(buf,Cor_parseEnv->parseptr,(unsigned int) n); \
	printf("buf = %s\n", buf);\
	printf("curr_buff = %d\n", Cor_parseEnv->curr_buff);\
        Cor_parseEnv->parseptr += n; \
        result = n; \
     }\
     else \
       result = YY_NULL; \
  }

#    else    // not STRING_PARSE

    // Scan input from a file pointer (default)
#define YY_INPUT(buf,result,max_size) \
    { \
        if (! (result = yyin_strm->gets(buf,max_size))) { \
           result = YY_NULL; \
        } \
    }

/*
        } else { \
            cout << "got: " << buf << endl; \
*/
#    endif    // not STRING_PARSE


EXTERN int yylex ARGS((YYSTYPE *lvalp));

#undef YYLMAX
#define YYLMAX 2048

/*  To give some args to yylex */
#undef YY_DECL
#define YY_DECL int yylex YY_PROTO((YYSTYPE *lvalp))

/**** Kludge to support Dont Care terms.
<INITIAL>"_"    { C_STRCAT(corparseEnv->commandbuf,yytext); return UNDERSCORE; }
***/


#define C_STRCAT(A,B) \
        if (Cor_exEnv->C_do_history_default) { \
            int tmp; \
            if ((tmp = Cor_parseEnv->command_offset + yyleng) \
                       < MAX_RULE_LENGTH) { \
                strncat((A)+Cor_parseEnv->command_offset,(B),yyleng); \
                Cor_parseEnv->command_offset = tmp ; \
            } else {\
                Cor_exEnv->C_do_history_default = 0;\
            }\
        }

%}
%Start C_COMMENT C_STR
%a 9000
%o 9000

dot             \.
digit           [0-9]
alpha           [a-zA-Z_]
capalpha        [A-Z]
alphanum        {alpha}|{digit}
optsign         [\+\-]?
litquote        \\\"

number          {digit}+
integer         {optsign}{number}
frac            {dot}{number}
mantissa        {number}|{number}{frac}|{frac}
exponent        [Ee]{integer}
real            {mantissa}{exponent}?

ssn_strt        {number}-{number}
ssn             {ssn_strt}+-{number}
date            {number}"/"{number}"/"{number}

%%

<INITIAL>\"[^\\\"]*\" {   /* Simple quoted strings */
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        yytext[yyleng-1] = '\0';
                        lvalp->carg = new Cor_CArg(yytext+1);
                        /*  corparseEnv->command_offset -= 2;  */
                        return SYMBOL;
                      }

<INITIAL>{ssn}      {   /* A special hack, only for table files. */
                        if (! Cor_reading_table)
                            REJECT;
 
                        lvalp->carg = new Cor_CArg(yytext);
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return SYMBOL;
                    }

<INITIAL>{date}     {   /* A special hack, only for table files. */
                        if (! Cor_reading_table)
                            REJECT;

                        lvalp->carg = new Cor_CArg(yytext);
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return SYMBOL;
                    }


<INITIAL>\"         { /* For strings that contain escaped characters */
                      C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                      BEGIN(C_STR);
                      str_ptr = str_buff;
                    }

<INITIAL>"cons"     { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return CONS; }
<INITIAL>"cons("    {
                      // This rule is for the benefit of the mini-parser
                      // which must avoid a look-ahead token
                        // put back the '('
                      yyless(yyleng-1);
                      C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                      return CONS;
                    }
<INITIAL>"nil"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return NIL; }

<INITIAL>"sips"     { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return SIPS; }
<INITIAL>"alias"    { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return ALIAS; }
<INITIAL>"mod"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return MOD; }
<INITIAL>"not"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return NOT; }
<INITIAL>"del"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return DEL; }

<INITIAL>"c_module" { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                      return START_C_MODULE;
                    }
<INITIAL>"end_c_module" { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                          return END_C_MODULE;
                        }
<INITIAL>"module"     { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return MODULE;
                      }
<INITIAL>"export"     { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return EXPORT;
                      }
<INITIAL>"end_module" { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return END_MODULE;
                      }

<INITIAL>"<<"       { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                      return LSHIFT;
                    }
<INITIAL>">>"       { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                      return RSHIFT;
                    }
<INITIAL>"bitxor"   { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                      return BITXOR;
                    }
<INITIAL>"bitand"   { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                      return BITAND;
                    }
<INITIAL>"bitor"    { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return BITOR; }
<INITIAL>"bitnot"   { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                      return BITNOT;
                    }  
<INITIAL>"is"       { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return EQUAL; }

<INITIAL>[A-Z]{alphanum}*  |
<INITIAL>[A-Z][A-Za-z0-9_.]*{alphanum}+    {
                            lvalp->carg = new Cor_CArg(yytext);
                            C_STRCAT(Cor_parseEnv->commandbuf,yytext);

                            // Table files cannot have variables
                            if (Cor_reading_table)
                                return SYMBOL;
                            else
                                return VAR;
                    }

<INITIAL>"_"        {
                        lvalp->carg = new Cor_CArg(yytext);
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return VAR;
                    }

<INITIAL>{number}   {
                        lvalp->carg = new Cor_CArg( atoi(yytext));
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext); 
                        return NUM;
                    }

<INITIAL>{real}     {
                        lvalp->carg = new Cor_CArg(atof(yytext));
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return NUM;
                    }

<INITIAL>[a-z]([#A-Za-z0-9_.]*[#A-Za-z0-9_])?"(" {
                        // This rule is for the benefit of the mini-parser
                        // which must avoid a look-ahead token
                          // put back the '('
                        yyless(yyleng-1);
                        lvalp->carg = new Cor_CArg(yytext);
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return FSYMBOL;
                    }

<INITIAL>[a-z]([#A-Za-z0-9_.]*[#A-Za-z0-9_])? {
                        lvalp->carg = new Cor_CArg(yytext);
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        return SYMBOL;
                    }


<INITIAL>","      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return COMMA; } 
<INITIAL>":"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return COLON; }
<INITIAL>"."      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return PERIOD;}
<INITIAL>":-"     { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return MARK; }
<INITIAL>"<-"     { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return MARK; }
<INITIAL>"+="     { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return APPEND;}
<INITIAL>"-="     { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return DELETE;}
<INITIAL>"*="     { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return UPDATE;}
<INITIAL>":="     { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return ASSIGN;}
<INITIAL>"==>"    { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return ARROW; }
<INITIAL>"//"     { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                    return SLASHSLASH;
                  }
<INITIAL>"("      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return LPAREN; }
<INITIAL>")"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return RPAREN; }
<INITIAL>"{"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return LBRACE; }
<INITIAL>"}"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return RBRACE; }
<INITIAL>"["      { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                    return LBRACKET;
                  }
<INITIAL>"]"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                    return RBRACKET;
                  }
<INITIAL>"$"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return DOLLAR; }
<INITIAL>"|"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return VBAR; }
<INITIAL>"@@"     {   // This is only for the reading of Table files (*.T)
                    return(DOUB_AT);
                  }
<INITIAL>"@"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return AT; }
<INITIAL>"+"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return PLUS; }
<INITIAL>"-"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return MINUS; }
<INITIAL>"*"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return STAR; }
<INITIAL>"/"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return SLASH; }
<INITIAL>"?"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                    return QUESTION;
                  }
<INITIAL>"="      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return EQUAL; }
<INITIAL>"\\="    { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return NOTEQUAL; }
<INITIAL>"~="     { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                    return NOTEQUAL;
                  }
<INITIAL>"<"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return LESS; }
<INITIAL>">"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return GREATER; }
<INITIAL>"<>"     { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                    return NOTEQUAL;
                  }
<INITIAL>"<="     { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                    return LESSEQUAL;
                  }
<INITIAL>">="     { C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                    return GREATEREQUAL;
                  }
<INITIAL>"~"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return NOT; }
<INITIAL>"!"      { C_STRCAT(Cor_parseEnv->commandbuf,yytext); return CUT; }




<INITIAL>"%".*$         ;    /* comment; ignore */
<INITIAL>"/*"           { BEGIN(C_COMMENT); Cor_parseEnv->in_comment = 1;}
<C_COMMENT>[^*\n]*      ;      /* ignore anything that's not a '*' */
<C_COMMENT>"*"+[^*/\n]* ;     /* ignore '*'s not followed by '/'s */
<C_COMMENT>\n           { Cor_parseEnv->cur_line_num++;
                          Cor_parseEnv->linebuf[0] = '\0';
                          Cor_parseEnv->line_offset = 0;
                        }

<C_COMMENT>"*"+"/"      { BEGIN(INITIAL); Cor_parseEnv->in_comment = 0;}


<C_STR>\"           { /* Closing quote */
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        *str_ptr = '\0';
                        lvalp->carg =  new Cor_CArg(str_buff);
                        BEGIN(INITIAL);
                        return SYMBOL;
                    }

<C_STR>[^\\\"]+     { /* Normal chars */
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        if ((str_ptr + yyleng) < str_end){
                            strcpy(str_ptr,yytext);
                            str_ptr += yyleng;
                        }
                    }

<C_STR>\\n          { /* Newline char */
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = '\n';
                    }

<C_STR>\\t          { /* Tab char */
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = '\t';
                    }

<C_STR>\\b          { /* Backspace char */
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = '\b';
                    }

<C_STR>\\f          { /* Formfeed char */
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = '\f';
                    }

<C_STR>\\.          { /* Literal char (like " and \) */
                        C_STRCAT(Cor_parseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = yytext[1];
                    }

[ \t\f]             ;     /* space/tab/formfeed; ignore */

\n|\r               {
                      Cor_parseEnv->cur_line_num++;
                      Cor_parseEnv->linebuf[0] = '\0';
                      Cor_parseEnv->line_offset = 0;
                    }

.                   {      /* any other single char; error */
                      *(Cor_exEnv->err_strm)
                          << "Error: Line " << Cor_parseEnv->cur_line_num
                          << ": Illegal Character";
                      if (yytext[0]>=' ' && yytext[0] <='~')
                        *(Cor_exEnv->err_strm) << ": `" << yytext[0] << "'\n";
                      else
                        *(Cor_exEnv->err_strm)
                                << ", ASCII: " << oct << setw(3) 
                                << ((int) yytext[0]) << " (octal)\n" << dec;
                    }

%%

//***********************************************************************
//
// yywrap() - This function is called by the lexer [yylex() or sslex()] when
//            the end-of-file (or end-of-string for sslex()) is reached.
//            It gets the next file/string ready (if any) and returns 1
//            which indicates to the lexer that there are no more files
//            remaining.  The lexer returns 0 to the parser, indicating
//            no more tokens are remaining. The parser function, yyparse(),
//            then returns 0, indicating that the parse is done.  This
//            behaviour allows a single item to be parsed to be placed
//            on the stack and parsed without the entire stack being consumed.
//
//***********************************************************************

int yywrap()
{
    if (Cor_parseEnv->in_comment) {
        *(Cor_exEnv->err_strm)
            << "ERROR: End of file reached inside comment.\n";
        Cor_parseEnv->in_comment = 0;
    }

    // Prepare the next file to be read, if there is one, but return
    // so that yyparse() returns in any case.

    Cor_parseStack->get_next_file();
    return(1);
}

//***********************************************************************
