// *************************************************************************
//    Profile.C
// *************************************************************************

#include <stdio.h>
#include <string.h>

#include "Profile.h"
#include "Config.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************

#define    COR_U_SEC 1000000

#ifdef    DO_PROFILE
Cor_ProfileInfo    Cor_GlobalProfileInfo;
#endif

// *************************************************************************
void Cor_ProfileInfo::clear()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_PROFILE
    i_unsuccesful = 0;
    i_succesful = 0;
    i_newfacts = 0;
    getnexts = 0;
    deleted_tuples = 0;
    non_unifying_tuples = 0;
    unifies = 0;
    Cor_subsumes = 0;
    invocations = 0;
    usage_time = 0.0;
#endif
}

// *************************************************************************
void Cor_ProfileInfo::init_rule()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_PROFILE
    invocations++;

    long t1=0, t2=0;

#ifndef    SOLAR_386
    struct rusage tmp_usage;
     
#    if    defined(SUN) || defined(LINUX)
    getrusage(RUSAGE_SELF, &tmp_usage);
#    endif

    t1 = tmp_usage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec;
#endif
    usage_time_st = ((double) t1) + ((double) t2) / COR_U_SEC;
#endif
}

// *************************************************************************
void Cor_ProfileInfo::exit_rule()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_PROFILE
    long t1=0, t2=0;

#ifndef    SOLAR_386
    struct rusage tmp_usage;
     
#    if    defined(SUN) || defined(LINUX)
    getrusage(RUSAGE_SELF, &tmp_usage);
#    endif

    t1 = tmp_usage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec;
#endif
    usage_time = (((double) t1) + ((double) t2) / COR_U_SEC) - usage_time_st;
#endif
}

// *************************************************************************
#ifdef    DO_PROFILE
void Cor_ProfileInfo::print(ostream *fd, char *str, int *predOrders,
    int numpreds) const
#else
void Cor_ProfileInfo::print(ostream*, char*, int*, int) const
#endif
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_PROFILE
    *fd << "_____" << str << "__PROFILE INFO_____\n";
    *fd <<  rule << "\n";
    if (numpreds && predOrders) {
        *fd << "Evaluation order : ";
        for (int i = 0; i < numpreds; i++)
            *fd <<  predOrders[i] << " ";
        *fd << endl;
    }
     
    *fd << "Invocations : " << invocations << ",    Usage time :  "
        << usage_time;
    *fd << "Instantiations : (Successful) " << i_succesful
        << ",   (Unsuccessful) " << i_unsuccesful << "\t|  ";

    *fd << "New Facts " << i_newfacts << " \n";
    *fd << "Unifies " << unifies << "\t| Subsumes " << Cor_subsumes
        << "\t| Non Unifying tuples " << non_unifying_tuples << endl;
    *fd << "GetNexts " << getnexts << "\t| Deleted Tuples "
        << deleted_tuples << endl << endl;
#endif
}

// *************************************************************************
#ifdef    DO_PROFILE
void Cor_AddProfileInfo(Cor_ProfileInfo *A, Cor_ProfileInfo *B) 
#else
void Cor_AddProfileInfo(Cor_ProfileInfo *, Cor_ProfileInfo *) 
#endif
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_PROFILE
    B->deleted_tuples += A->deleted_tuples;
    B->non_unifying_tuples += A->non_unifying_tuples;
    B->i_succesful += A->i_succesful;
    B->i_unsuccesful += A->i_unsuccesful;
    B->i_newfacts += A->i_newfacts;
    B->getnexts    += A->getnexts;
    B->unifies += A->unifies;
    B->Cor_subsumes += A->Cor_subsumes;
    B->invocations += A->invocations;
    B->usage_time += A->usage_time;
#endif
}

// *************************************************************************
