// *************************************************************************
//    FileDescStruct.h
// *************************************************************************

#ifndef _FileDescStruct_H_
#define _FileDescStruct_H_

#include "sm_client.h"
#include "Debug.h"

// *************************************************************************
/*
    Structure of file descriptor object in catalog file on disk
    Each field of the relation could possibly have an index.
    This is sufficient because we are only considering ground
    and flat tuples on disk.
*/
// *************************************************************************

class Cor_FileDescStruct {

    friend class Cor_PersistentRel;

  public:
    Cor_FileDescStruct() {
                           INVALIDATE_FID(fid);
                           num_indexes = 0;
                           count = 0;
                           duplicates = 0;
                         }

    // Takes schema description of a relation as a functor parameter
    Cor_FileDescStruct( Cor_CArg *arg, int &error,
                        int temporary = 0, int by_example=0);

    FID    fid;
    Cor_BPlusTreeIndex indexes[COR_MAX_INDEXES];
    int num_indexes;
    Cor_Schema schema;
    int count;
    int duplicates;

    void copyDesc(Cor_FileDescStruct *);

    // Returns offset where an index entry can be made
    int canAddIndex();

    // Returns offset of the next valid index after offset i
    // The first time, this is called with i = -1
    int nextIndex(int i);

    // Adds an index at position 'offset' in the index array. The attributes
    // of the index are specified by buf. Duplicate indexes are _not_ created.
    // returns the offset of the desired index
    int addIndex(int offset, Cor_BitVector *bv);

    // Delete an index whose offset is specified 
    void destroyIndex(int offset);

    // Adds an object to a specified index
    void addObjectToIndex(int offset, USERDESC *u_desc);

    // Find an index that best matches the given attr string
    virtual int findBestMatch(char *attr_str);

    // Return the arity of the specified index
    int indexArity(int index_offset) {
        return indexes[index_offset].arity;
    };

    void print_index_adornments(ostream *outf);

    void startIndexScan(int index_offset) {
        indexes[index_offset].begin_load();
    }

    void endIndexScan(int index_offset) {
        indexes[index_offset].end_load();
    }
    void abortIndexScan(int index_offset) {
        indexes[index_offset].abort_load();
    }
};

// *************************************************************************

#endif _FileDescStruct_H_
