/*****************************************************************************/
//	Linearity.C
/*****************************************************************************/

#include "Linearity.h"

#include "Literal.h"

/*****************************************************************************/

static int right_linear(Cor_Clause * rule);
static int left_linear(Cor_Clause *rule);
static int test_multi_linear(Cor_Clause *rule, Cor_Literal *pred);
static int multi_linear(Cor_Clause * rule);

/*****************************************************************************/
int assign_linear_status(Cor_Clause *rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
	//	NOTE: Have to check that the head has all distinct variables.

//#	ifdef DEBUG
//		print_rule(rule, cerr_strm);
//#	endif

	if (!rule->recursive) {
		rule->factor_status = (int) COR_NON_RECURSIVE;
		rule->factor_pred = NULL;
		return 1;
	}

	if (right_linear(rule)) {
		return 1;
	}

	if (left_linear(rule)) {
		return 1;
	}

	if (multi_linear(rule)) {
		return 1;
	}

	rule->factor_status = (int) COR_NON_LINEAR;
	rule->factor_pred = NULL;
	return 0;
}

/*****************************************************************************/
static int right_linear(Cor_Clause *rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
	Cor_Literal *pred = NULL;

	//	Only one literal should have the same predicate as the head.
    int i;
	for (i = 0; i< rule->num_preds; i++) {
		if (rule->head->predicate == rule->preds[i]->predicate) {
			 //	If the head predicate has already occured in the rule.
			 if (pred) return 0;
			 pred = rule->preds[i];
		}
	}

	if (pred == NULL) return 0;

	//	This literal should have the same adornment as the head.
	if (!(rule->head->adorn == pred->adorn))
		return 0;

	//	Checking conditions on variables in the predicates!
	for (i=0; i< pred->adorn.len(); i++) {
		if (!(pred->adorn[i])) {
			// The FREE variables of pred should be identical to the free
			// variables of the head predicate.
			if (!(pred->args[i].equals(rule->head->args[i]))) return 0;

			// Check that the free variables do not occur anywhere else
			// in the rule.
            int j;
			for (j = 0; j < rule->num_preds; j++) {
				if (rule->preds[j] != pred) {
					for (int k = 0; k < rule->preds[j]->args.count(); k++) {
						if (pred->args[i].equals(rule->preds[j]->args[k]))
							return 0;
					}
				}
			}
		}
		else {
			// The BOUND variables of pred should appear in one of the
			// body literal or as a bound variable of the head literal.
			int var_found = 0;
            int j;

			for (j = 0; j < rule->head->adorn.len(); j++) {
				if ((pred->args[i].equals(rule->head->args[j])) &&
					(rule->head->adorn[j]))
				{
					var_found = 1;
					break;
				}
			}
			if (var_found) continue;

			for (j = 0; j < rule->num_preds; j++) {
				if (rule->preds[j] != pred) {
					for (int k = 0; k < rule->preds[j]->args.count(); k++) {
						if (pred->args[i].equals(rule->preds[j]->args[k])) {
							var_found = 1;
							break;
						}
					}
					if (var_found) break;
				}
			}
			if (!(var_found))  return 0;
		}
	}

	rule->factor_status = (int) COR_RT_LINEAR;
	rule->factor_pred = pred;
	return 1;
}

/*****************************************************************************/
static int left_linear(Cor_Clause *rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
	Cor_Literal *pred = NULL;
	int pseudo_linear = 0;

	// Only one literal should have the same predicate as the head.
    int i;
	for (i = 0; i< rule->num_preds; i++) {
		if (rule->head->predicate == rule->preds[i]->predicate) {
			//	If the head predicate has already occured in the rule.
			if (pred) return 0;
			pred = rule->preds[i];
		}
	}

	if (pred == NULL) return 0;

	//	This literal should have the same adornment as the head.
	if (!(rule->head->adorn == pred->adorn))
		return 0;

	//	Checking conditions on variables in the predicates!
	for (i=0; i< pred->adorn.len(); i++) {
		if (pred->adorn[i]) {
			// The BOUND variables of pred should be identical to the bound
			// variables of the head literal.
			if (!(pred->args[i].equals(rule->head->args[i])))
                return 0;

			//	Check that the bound variables do not occur anywhere else
			//	in the rule. If this condition is not satisfied, the rule
			//	could still be PSEUDO_LT_LINEAR.

			if (pseudo_linear) continue;
			for (int j = 0; j < rule->num_preds; j++) {
				if (rule->preds[j] != pred) {
					for (int k = 0; k < rule->preds[j]->args.count(); k++) {
						if (pred->args[i].equals(rule->preds[j]->args[k])) {
							pseudo_linear = 1;
							break;
						}
					}
					if (pseudo_linear) break;
				}
			}
		}
		else {
			//	The FREE variables of pred should be different from all
			//	bound variables of pred (i.e) of the head literal.

			for (int j = 0; j < rule->head->adorn.len(); j++) {
				if ((pred->args[i].equals(rule->head->args[j])) &&
					(rule->head->adorn[j]))
				{
					return 0;
				}
			}
		}
	}

	
	rule->factor_status = (pseudo_linear) ? (int) PSEUDO_LT_LINEAR :
		(int) COR_LT_LINEAR;
	rule->factor_pred = pred;
	return 1;
}

/*****************************************************************************/
static int multi_linear(Cor_Clause * rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
	Cor_Literal *pred = NULL, *the_pred = NULL;
	int preds_found = 0;
    int i;
	for (i = 0; i< rule->num_preds; i++) {
		pred = rule->preds[i];
		if (rule->head->predicate == pred->predicate) {
			// The predicate is the same as the head predicate.
			preds_found++;

			if (test_multi_linear(rule, pred)) {
				// The literal satisfies the COR_MULTI_LINEAR requirements.
				if (the_pred) return 0;
				the_pred = pred;
			}
			else {
				// The literal has the same predicate as the head but does
				// not satisfy the COR_MULTI_LINEAR requirements.
                int k;
				for (k=0; k< pred->adorn.len(); k++) {
					if (rule->head->adorn[k]) {
						// The BOUND variables should be identical to the head.
						if (!(pred->args[k].equals(rule->head->args[k])))
							return 0;
					}
					else {
						// Any other variables are disjoint from the bound
						// variables of the head literal.
						for (int j = 0; j < rule->head->adorn.len(); j++) {
							if ((pred->args[k].equals(rule->head->args[j])) &&
								(rule->head->adorn[j])) return 0;
						}
					}
				}
			}
		}
		else {
			// The predicate is different from the head predicate.  All its
			// variables ahould be disjoint from the bound variables of the
			// head literal.
			for (int k=0; k < pred->adorn.len(); k++) {
				for (int j = 0; j < rule->head->adorn.len(); j++) {
					if ((pred->args[k].equals(rule->head->args[j])) &&
						(rule->head->adorn[j])) return 0;
				}
			}
		}
	}
	
	if (preds_found < 2 || !the_pred) return 0;
	
	rule->factor_status = (int) COR_MULTI_LINEAR;
	rule->factor_pred = the_pred;
	return 1;
}

/*****************************************************************************/
/*
	Tests to see if the particular literal 'pred' is the one on which
	the multi-linear conditions hold.
*/

static int test_multi_linear(Cor_Clause *rule, Cor_Literal *pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
	//	The adornment should be the same as for the head literal.
	if (!(rule->head->adorn == pred->adorn))
		return 0;

	for (int i=0; i< pred->adorn.len(); i++) {
		if (pred->adorn[i]) {
			//	The BOUND variables of pred should appear in one of the
			//	body literal but NOT as a bound variable of the head literal.
			int var_found = 0;
            int j;
			for (j = 0; j < rule->head->adorn.len(); j++) {
				if ((pred->args[i].equals(rule->head->args[j])) &&
					(rule->head->adorn[j])) return 0;
			}

			for (j = 0; j < rule->num_preds; j++) {
				if (rule->preds[j] != pred) {
				    for (int k = 0; k < rule->preds[j]->args.count(); k++) {
						if (pred->args[i].equals(rule->preds[j]->args[k])) {
							var_found = 1;
							break;
						}
				    }
				    if (var_found) break;
				}
			}
		    if (!(var_found))  return 0;
		    
		}
		else {
			//	The FREE variables of pred should be identical to the free
			//	variables of the head predicate.
			if (!(pred->args[i].equals(rule->head->args[i]))) return 0;

			//	Check that the free variables do not occur anywhere else in
			//	the rule.
			for (int j = 0; j < rule->num_preds; j++) {
				if (rule->preds[j] != pred) {
					for (int k = 0; k < rule->preds[j]->args.count(); k++) {
						if (pred->args[i].equals(rule->preds[j]->args[k]))
							return 0;
					}
				}
			}
		}
	}
	return 1;
}

/*****************************************************************************/

